package gs.config.v0.audit

import cats.effect.Ref
import cats.effect.Sync
import cats.syntax.all.*
import gs.config.v0.ConfigName

trait ConfigManifest[F[_]]:
  /** Retrieve a snapshot of the current state of this configuration manifest.
    * This state tracks all configuration names that the caller attempted to
    * access, along with a record of whether each attempt succeeded or failed.
    *
    * @return
    *   The current state of this manifest.
    */
  def snapshot(): F[Map[ConfigName, List[ConfigQueryResult]]]

  /** Record a query for some [[ConfigName]] in this manifest.
    *
    * @param name
    *   The [[ConfigName]] that was queried.
    * @param queryResult
    *   The [[ConfigQueryResult]] describing the result.
    * @return
    *   Side-effect indicating that the query was recorded.
    */
  def record(
    name: ConfigName,
    queryResult: ConfigQueryResult
  ): F[Unit]

object ConfigManifest:

  /** Instantiate a new, empty, standard manifest.
    */
  def standard[F[_]: Sync]: F[ConfigManifest[F]] =
    Standard.initialize[F]

  /** Standard implementation of [[ConfigManifest]]. Collects an in-memory
    * collection to audit configuration access.
    *
    * @param manifest
    *   The underlying manifest.
    */
  final class Standard[F[_]: Sync] private (
    private val manifest: Ref[F, Map[ConfigName, List[ConfigQueryResult]]]
  ) extends ConfigManifest[F]:

    override def snapshot(): F[Map[ConfigName, List[ConfigQueryResult]]] =
      manifest.get

    override def record(
      name: ConfigName,
      query: ConfigQueryResult
    ): F[Unit] =
      manifest.update(addQueryToName(name, query, _))

    private def addQueryToName(
      name: ConfigName,
      query: ConfigQueryResult,
      state: Map[ConfigName, List[ConfigQueryResult]]
    ): Map[ConfigName, List[ConfigQueryResult]] =
      state.get(name) match
        case Some(queries) => state.updated(name, queries ++ List(query))
        case None          => state + (name -> List(query))

  object Standard:

    /** Instantiate a new, empty, standard manifest.
      */
    def initialize[F[_]: Sync]: F[ConfigManifest[F]] =
      Ref
        .of(Map.empty[ConfigName, List[ConfigQueryResult]])
        .map(m => new Standard[F](m))

  end Standard

end ConfigManifest
